"""
Pattern association network demonstration program

Please do not redistribute this code or modifications of it without the following notice:
(c) E T Rolls www.oxcns.org 2015
The background to these networks is provided in
Rolls ET (2021) Brain Computations: A Systems-Level Analysis (Oxford University Press) Appendix 2 (B)
which is available at http://www.oxcns.org by following the link to the book.
The instructions and exercises for this code are in
Rolls ET (2021) Brain Computations: A Systems-Level Analysis (Oxford University Press) Appendix 4 (D)
which is available at http://www.oxcns.org by following the link to the book.

This program is a translation into Python of the Matlab / Octave program PatternAssocDemo.m.
This program is intended to demonstrate some of the principles of how pattern association networks operate,
and for that reason code optimizations are frequently eschewed in favour of illustrating principles of neuronal operation.
This program therefore also uses few external libraries.

Note that if the program stops, it may be waiting for keyboard input.

The synaptic weight matrix (SynMat) has column vectors to represent the synapses on a single neuron.
There is one column vector therefore for each of the N neurons.
Thus, SynMat[row, column] implements SynMat[synapse, neuron]
"""

# import necessary libraries
import numpy as np
import matplotlib.pyplot as plt
import warnings
warnings.filterwarnings("ignore")   # do not show unnecessary figure axis warnings


fig = 1                                     # index of plotting figures
N = 8                                       # number of neurons in the pattern associator
nSyn = 64                                   # number of synapses on each neuron

# parameters for some training patterns that overlap by shift
leng = 16                                   # length of each training pattern
InputSparseness = leng / nSyn               # the sparseness of the input CS patterns is the number of bits on for any one pattern
shift = 8                                   # shift between training patterns
nPatts = 8                                  # number of training patterns
Learnrate = 1 / nPatts                      # not of especial significance in an autoassociation network, but this keeps the weights wiehin a range
OutputSparseness = 1 / N                    # The US patterns have one neuron on in this simple simulation
display = 1                                 # 0 no display; 1 display the network
nFlipBits = 8                               # The number of bits that are flipped to produce distorted recall cues


# make some training patterns
TrainPatts = np.zeros((nSyn, nPatts))       # the conditioned stimuli
for patt in range(nPatts):
    for x in range(leng):
        syn = patt * shift + x
        syn = syn % nSyn
        TrainPatts[syn][patt] = 1


# TrainPatts
if display == 1:
    fig += 1
    plt.figure(fig)
    plt.title('Training Patterns')
    plt.xlabel('CS Pattern Number')
    plt.ylabel('CS Training Pattern')
    plt.imshow(TrainPatts, cmap='gray', aspect='auto')
    #plt.show()
    plt.pause(1)



# Now we make some testing patterns with some bits flipped from the training patterns
TrainPattsFlipped = np.copy(TrainPatts)
for patt in range(nPatts):
    synarray = np.random.permutation(range(nSyn))
    el = 0
    for bit in range(nFlipBits):
        while TrainPatts[synarray[el], patt] != 1:
            el += 1
            if el > nSyn:
                print('Error: too many bits being flipped')
                el = 1
        TrainPattsFlipped[synarray[el], patt] = 0
        el += 1

    synarray = np.random.permutation(range(nSyn))
    el = 0
    for bit in range(nFlipBits):
        while TrainPatts[synarray[el], patt] != 0:
            el += 1
            if el > nSyn:
                print('Error: too many bits being flipped')
                el = 1
        TrainPattsFlipped[synarray[el], patt] = 1
        el += 1


if display == 1:
    fig += 1
    plt.figure(fig)
    plt.title('Training patterns with ' + str(nFlipBits) + ' flipped bits for testing')
    plt.xlabel('CS Pattern Number')
    plt.ylabel('CS distorted testing pattern')
    plt.imshow(TrainPattsFlipped, cmap='gray', aspect='auto')
    #plt.show()
    plt.pause(1)


SynMat = np.zeros((nSyn, N))
Actvn = np.zeros(N)                    # activation
Rate = np.zeros(N)                     # output firing rates
US = np.zeros((nPatts, N))                  # unconditioned stimulus, also the target for retrieval
for patt in range(nPatts):
    US[patt, patt] = 1


# Train. See Rolls (2021) B.2.7.1 for why an associative learning rule with heterosynaptic long term depression is useful
for patt in range(nPatts):
    for neuron in range(N):
        for syn in range(nSyn):
            preSynRate = TrainPatts[syn, patt]          # the CS patterns used for training
            postSynRate = US[patt, neuron]
            # weight_change = postSynRate * preSynRate; % an associative, Hebbian, synaptic modification rule
            weight_change = Learnrate * postSynRate * (preSynRate - InputSparseness) # OR use a Hebb rule but with also heterosynaptic LTD see Rolls (2008) B.3.3.6.
            SynMat[syn, neuron] += weight_change


print('Synaptic matrix minimum =', np.min(SynMat), 'maximum=', np.max(SynMat))
# comment in the following two lines to test whether the network operates with positive-only synaptic weights
# SynMat = SynMat - np.min(SynMat)
# print('Synaptic matrix minimum =', np.min(SynMat), 'maximum=', np.max(SynMat))


# Test
print('Testing')
nCorrect = 0                        # number of correctly recalled patterns
for patt in range(nPatts):
    # calculate activation
    for neuron in range(N):
        # PreSynInput = TrainPatts(:, patt); # the retrieval cue is the conditioned stimulus
        PreSynInput = TrainPattsFlipped[:,patt]
        Actvn[neuron] = np.dot(PreSynInput, SynMat[:,neuron])

    scale = 1.0 / (np.max(Actvn) - np.min(Actvn))
    Actvn = (Actvn - np.min(Actvn)) * scale

    # now convert the activation into a firing rate using a binary or linear threshold activation function
    # the threshold is selected in this case artificially by sorting
    # the array to determine where the threshold should be set to achieve the required sparseness
    # in the brain, inhibitory neurons are import in setting the proportion of excitatory neurons that are active

    tmp = sorted(Actvn, reverse=True)
    el = np.floor(len(Rate) * OutputSparseness) - 1
    Threshold = tmp[int(el)]
    Rate = np.copy(Actvn)


    for neuron in range(len(Rate)):
        if Rate[neuron] >= Threshold: # threshold binary activation function
            # Rate[neuron] = 1 # comment out this line for a threshold linear activation function
            pass
        else:
            Rate[neuron] = 0


    if display == 1:
        plt.figure(101)
        ax2 = plt.axes(title='Synaptic Matrix', position=[0.2, 0.25, 0.7, 0.7])
        ax2.imshow(SynMat, cmap='gray', aspect='auto')
        ax3 = plt.axes(title='Activation', position=[0.2, 0.125, 0.7, 0.03])
        ax3.set_yticklabels([])
        ax3.imshow(np.reshape(Actvn, (1, -1)), cmap='gray', aspect='auto')
        ax4 = plt.axes(title='Postsynaptic Rate', position=[0.2, 0.0, 0.7, 0.03])
        ax4.set_yticklabels([])
        ax4.imshow(np.reshape(Rate, (1, -1)), cmap='gray', aspect='auto')
        ax1 = plt.axes(title='Presynaptic input', position=[0.1, 0.25, 0.02, 0.7])
        ax1.set_xticklabels([])
        ax1.imshow(np.reshape(PreSynInput, (-1,1)), cmap='gray', aspect='auto')
        #plt.show()
        plt.pause(1)
        input(f'Press any key to test with pattern {patt+2}: ')


    # evaluate performance
    R1 = np.corrcoef(US[:,patt], Rate) # correlationn coefficient
    R = R1[0, 1]
    #print('Patt', str(patt+2), 'correlation of recalled firing rate vector with training pattern = ', str(R))
    print(f'Patt {patt+2} correlation of recalled firing rate vector with training pattern = {R:0.3f}')
    if R > 0.98:
        nCorrect += 1

PercentCorrect = nCorrect / nPatts * 100;
print('Percentage correct recall =', str(PercentCorrect))
